"""Module inspection with Streamlit."""
import logging
import sys

import nnrl.utils as ptu
import numpy as np
import pandas as pd
import ray
import streamlit as st
import torch
from bokeh.layouts import gridplot
from bokeh.models import ColumnDataSource, DataRange1d
from bokeh.plotting import figure
from ray.rllib import SampleBatch

import raylab
import raylab.utils.dictionaries as dutil
from raylab.utils.checkpoints import get_agent_from_checkpoint

# pylint:disable=invalid-name,missing-docstring,pointless-string-statement
# pylint:disable=no-value-for-parameter
"""
# Visualize Module
"""

logging.getLogger("ray.rllib").setLevel(logging.DEBUG)


@st.cache
def setup():
    ray.init(ignore_reinit_error=True)
    raylab.register_all_agents()
    raylab.register_all_environments()


@st.cache
def load_agent(checkpoint, agent_id, evaluate):
    return get_agent_from_checkpoint(
        checkpoint,
        agent_id,
        use_eval_config=evaluate,
        config_overrides={
            "num_workers": 0,
            "num_envs_per_worker": 1,
            "batch_mode": "complete_episodes",
            "rollout_fragment_length": 1,
        },
    )


def produce_rollout(agent):
    return agent.workers.local_worker().sample()


def scatter_matrix(dataset, target_dataset=None):
    # pylint:disable=too-many-locals
    # https://stackoverflow.com/a/49634658
    dataset_source = ColumnDataSource(data=dataset)

    xdr = DataRange1d(bounds=None)
    ydr = DataRange1d(bounds=None)

    columns = dataset.columns
    if target_dataset is not None:
        assert len(target_dataset.columns) == len(columns)
        target_source = ColumnDataSource(data=target_dataset)

    y_max = len(columns)
    scatter_plots = []
    for yidx, y_col in enumerate(columns):
        scatter_plots += [[]]
        scatter_plots[yidx] += [None] * yidx
        for xidx in range(yidx, y_max):
            x_col = columns[xidx]

            pic = figure(
                x_range=xdr,
                y_range=ydr,
                plot_width=200,
                plot_height=200,
                min_border_left=2,
                min_border_right=2,
                min_border_top=2,
                min_border_bottom=2,
                tools="",
            )
            rend = pic.circle(
                source=dataset_source,
                x=x_col,
                y=y_col,
                fill_alpha=0.3,
                line_alpha=0.3,
                size=3,
            )
            # pylint:disable=no-member
            xdr.renderers.append(rend)
            ydr.renderers.append(rend)
            # pylint:enable=no-member
            if target_dataset is not None:
                rend = pic.circle(
                    source=target_source,
                    x=target_dataset.columns[xidx],
                    y=target_dataset.columns[yidx],
                    fill_alpha=0.3,
                    line_alpha=0.3,
                    size=10,
                    color="red",
                )
                # pylint:disable=no-member
                xdr.renderers.append(rend)
                ydr.renderers.append(rend)
                # pylint:enable=no-member

            pic.yaxis.axis_label = ""
            pic.yaxis.visible = False
            pic.xaxis.axis_label = ""
            pic.xaxis.visible = False

            scatter_plots[yidx] += [pic]

    return gridplot(scatter_plots, sizing_mode="scale_both")


def make_histograms(dataset, bins, ranges=()):
    columns = dataset.columns
    limits = tuple(zip(*ranges)) if ranges else (None,) * len(columns)

    xdr = DataRange1d(bounds=None)
    # ydr = DataRange1d(bounds=None)
    # ydr.start = 0

    row = []
    for col, limit in zip(columns, limits):
        hist, edges = np.histogram(dataset[col], density=True, bins=bins, range=limit)
        pic = figure(
            title=col,
            x_range=xdr,
            # y_range=ydr,
            plot_width=200,
            plot_height=240,
            min_border_left=2,
            min_border_right=2,
            min_border_top=2,
            min_border_bottom=42,
            tools="",
        )
        pic.y_range.start = 0
        rend = pic.quad(
            top=hist,
            bottom=0,
            left=edges[:-1],
            right=edges[1:],
            fill_color="navy",
            line_color="white",
            alpha=0.5,
        )
        # pylint:disable=no-member
        xdr.renderers.append(rend)
        # ydr.renderers.append(rend)
        # pylint:enable=no-member

        pic.grid.grid_line_color = "white"
        pic.yaxis.axis_label = ""
        pic.yaxis.visible = False
        row += [pic]

    return gridplot([row], sizing_mode="scale_width")


@st.cache(
    max_entries=1,
    # pylint:disable=protected-access
    hash_funcs={torch.jit.RecursiveScriptModule: hash},
    # pylint:enable=protected-access
)
def get_model_samples(module, row, n_samples):
    obs = ptu.convert_to_tensor(row[SampleBatch.CUR_OBS], "cpu")
    act = ptu.convert_to_tensor(row[SampleBatch.ACTIONS], "cpu")
    new_obs, _ = module.sample(obs, act, (n_samples,))
    return new_obs.detach().numpy()


def plot_model_distributions(policy, row):
    model = policy.module.model
    n_samples = st.number_input("Number of observation samples", min_value=1, step=1)
    samples = get_model_samples(model, row, n_samples)

    data = {f"obs[{i}]": samples[..., i] for i in range(samples.shape[-1])}
    dataset = pd.DataFrame(data)

    new_obs = row[SampleBatch.NEXT_OBS]
    target_data = {f"new_obs[{i}]": new_obs[i][None] for i in range(new_obs.shape[-1])}
    target_dataset = pd.DataFrame(target_data)

    bins = st.number_input(
        "Histogram bins",
        min_value=1,
        max_value=n_samples,
        value=(1 + n_samples) // 2,
        step=1,
    )
    st.bokeh_chart(make_histograms(dataset, bins))
    st.bokeh_chart(scatter_matrix(dataset, target_dataset))


@st.cache(
    max_entries=1,
    # pylint:disable=protected-access
    hash_funcs={
        torch.jit.RecursiveScriptModule: id,
        torch._C._TensorBase: id,
        torch.Tensor: id,
    },
    # pylint:enable=protected-access
)
def get_actor_outputs(module, row, n_samples):
    obs = ptu.convert_to_tensor(row[SampleBatch.CUR_OBS], "cpu")
    with torch.no_grad():
        acts, logp = module.sample(obs, (n_samples,))
        deterministic, _ = module.deterministic(obs)
        deterministic.unsqueeze_(0)
    log_prob = module.log_prob(obs, acts)
    entropy = -log_prob.mean()
    nll_grad = ptu.flat_grad(entropy, module.parameters())
    return {
        "acts": acts,
        "logp": logp,
        "det": deterministic,
        "log_prob": log_prob.detach(),
        "entropy": entropy.detach(),
        "nll_grad": nll_grad,
    }


def plot_action_distributions(outputs, bins, ranges=()):
    acts, det = map(lambda x: x.numpy(), dutil.get_keys(outputs, "acts", "det"))
    data = {f"act[{i}]": acts[..., i] for i in range(acts.shape[-1])}
    dataset = pd.DataFrame(data)

    det_data = {f"det[{i}]": det[..., i] for i in range(det.shape[-1])}
    det_dataset = pd.DataFrame(det_data)

    st.bokeh_chart(make_histograms(dataset, bins, ranges=ranges))
    st.bokeh_chart(scatter_matrix(dataset, det_dataset))


def plot_logp_stats(outputs, bins):
    st.write("Entropy:", outputs["entropy"])
    st.write("grad_norm(nll):", outputs["nll_grad"].norm(p=2))

    dataset = pd.DataFrame({k: outputs[k].numpy() for k in "logp log_prob".split()})
    st.bokeh_chart(make_histograms(dataset, bins))
    dataset = pd.DataFrame({"nll_grad": outputs["nll_grad"].numpy()})
    st.bokeh_chart(make_histograms(dataset, bins))


def viz_actor_distributions(policy, row):
    actor = policy.module.actor
    n_samples = st.number_input("Number of action samples", min_value=1, step=1)

    outputs = get_actor_outputs(actor, row, n_samples)

    bins = st.number_input(
        "Histogram bins",
        min_value=1,
        max_value=n_samples,
        value=(1 + n_samples) // 2,
        step=1,
    )
    plot_action_distributions(
        outputs, bins, ranges=(policy.action_space.low, policy.action_space.high)
    )
    plot_logp_stats(outputs, bins)


def viz_distributions(policy, rollout, timestep):
    row = list(rollout.rows())[timestep]

    components = []
    if "actor" in policy.module:
        components += ["actor"]
    if "model" in policy.module:
        components += ["model"]

    component = st.selectbox("Inspect module:", options=components)
    if component == "actor":
        viz_actor_distributions(policy, row)
    elif component == "model":
        plot_model_distributions(policy, row)


# https://discuss.streamlit.io/t/how-can-i-clear-a-specific-cache-only/1963/6
@st.cache(allow_output_mutation=True)
def get_rollout(agent):
    return [produce_rollout(agent)]


def main():
    setup()
    agent_id, checkpoint = sys.argv[1], sys.argv[2]

    evaluate = st.checkbox("Use evaluation config")
    agent = load_agent(checkpoint, agent_id, evaluate)

    rollout_wrapper = get_rollout(agent)
    if st.button("Sample new episode"):
        rollout_wrapper.clear()
        rollout_wrapper.append(produce_rollout(agent))

    rollout = rollout_wrapper[0]
    timestep = st.slider("Timestep", min_value=0, max_value=rollout.count - 1)

    viz_distributions(agent.get_policy(), rollout, timestep)


if __name__ == "__main__":
    main()



"""Base Policy with common methods for all SVG variations."""
import torch
from nnrl.nn.utils import update_polyak
from ray.rllib import SampleBatch

from raylab.options import configure, option
from raylab.policy import EnvFnMixin, TorchPolicy
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.policy.losses import ISFittedVIteration, MaximumLikelihood


@configure
@option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@option("max_is_ratio", 5.0, help="Clip importance sampling weights by this value")
@option(
    "vf_loss_coeff",
    1.0,
    help="Weight of the fitted V loss in the joint model-value loss",
)
class SVGTorchPolicy(EnvFnMixin, TorchPolicy):
    """Stochastic Value Gradients policy using PyTorch."""

    # pylint:disable=abstract-method
    dist_class = WrapStochasticPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        self.loss_model = MaximumLikelihood(self.module.model)
        self.loss_critic = ISFittedVIteration(
            self.module.critic, self.module.target_critic
        )
        self.loss_critic.gamma = self.config["gamma"]

    @torch.no_grad()
    def add_truncated_importance_sampling_ratios(self, batch_tensors):
        """Compute and add truncated importance sampling ratios to tensor batch."""
        is_ratios = self.importance_sampling_ratios(batch_tensors)
        _is_ratios = torch.clamp(is_ratios, max=self.config["max_is_ratio"])
        batch_tensors[ISFittedVIteration.IS_RATIOS] = _is_ratios
        return batch_tensors, {"is_ratio_mean": is_ratios.mean().item()}

    def importance_sampling_ratios(self, batch_tensors):
        """Compute unrestricted importance sampling ratios."""
        curr_logp = self.module.actor.log_prob(
            batch_tensors[SampleBatch.CUR_OBS], batch_tensors[SampleBatch.ACTIONS]
        )
        is_ratio = torch.exp(curr_logp - batch_tensors[SampleBatch.ACTION_LOGP])
        return is_ratio

    def compute_joint_model_value_loss(self, batch_tensors):
        """Compute model MLE loss and fitted value function loss."""
        mle_loss, mle_info = self.loss_model(batch_tensors)
        isfv_loss, isfv_info = self.loss_critic(batch_tensors)

        loss = mle_loss + self.config["vf_loss_coeff"] * isfv_loss
        return loss, {**mle_info, **isfv_info}

    def _update_polyak(self):
        update_polyak(
            self.module.critic, self.module.target_critic, self.config["polyak"]
        )



"""NAF policy class using PyTorch."""
import torch
from nnrl.nn.critic import ClippedVValue
from nnrl.nn.utils import update_polyak
from nnrl.optim import build_optimizer
from nnrl.types import TensorDict
from ray.rllib.utils import override
from torch import nn

from raylab.options import configure, option
from raylab.policy import TorchPolicy
from raylab.policy.action_dist import WrapDeterministicPolicy
from raylab.policy.losses import FittedQLearning
from raylab.policy.off_policy import OffPolicyMixin, off_policy_options


@configure
@off_policy_options
@option("optimizer/type", "Adam")
@option("optimizer/lr", 3e-4)
@option(
    "polyak",
    0.995,
    help="Interpolation factor in polyak averaging for target networks.",
)
@option("module/type", "NAF")
@option("module/separate_behavior", True)
@option("exploration_config/type", "raylab.utils.exploration.ParameterNoise")
@option(
    "exploration_config/param_noise_spec",
    {"initial_stddev": 0.1, "desired_action_stddev": 0.2, "adaptation_coeff": 1.01},
)
@option("exploration_config/pure_exploration_steps", 1000)
class NAFTorchPolicy(OffPolicyMixin, TorchPolicy):
    """Normalized Advantage Function policy in Pytorch to use with RLlib."""

    # pylint:disable=abstract-method
    dist_class = WrapDeterministicPolicy

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_fn = FittedQLearning(
            self.module.critics, ClippedVValue(self.module.target_vcritics)
        )
        self.loss_fn.gamma = self.config["gamma"]

        self.build_replay_buffer()

    @override(TorchPolicy)
    def _make_module(self, obs_space, action_space, config):
        module_config = config["module"]
        module_config["type"] = "NAF"
        # pylint:disable=no-member
        return super()._make_module(obs_space, action_space, config)

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        optimizers.update(
            naf=build_optimizer(self.module.critics, self.config["optimizer"])
        )
        return optimizers

    @override(OffPolicyMixin)
    def improve_policy(self, batch: TensorDict) -> dict:
        with self.optimizers.optimize("naf"):
            loss, info = self.loss_fn(batch)
            loss.backward()

        info.update(self.extra_grad_info())

        vcritics, target_vcritics = self.module.vcritics, self.module.target_vcritics
        update_polyak(vcritics, target_vcritics, self.config["polyak"])
        return info

    @torch.no_grad()
    def extra_grad_info(self):
        """Compute gradient norm for components."""
        return {
            "grad_norm": nn.utils.clip_grad_norm_(
                self.module.critics.parameters(), float("inf")
            ).item()
        }



"""Trainer and configuration for SVG(1)."""
from raylab.agents import Trainer
from raylab.agents.model_based import set_policy_with_env_fn
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import SVGOneTorchPolicy


@configure
@OffPolicyMixin.add_options
class SVGOneTrainer(OffPolicyMixin, Trainer):
    """Single agent trainer for SVG(1)."""

    _name = "SVG(1)"
    _policy_class = SVGOneTorchPolicy

    def after_init(self):
        super().after_init()
        set_policy_with_env_fn(self.workers, fn_type="reward")



"""Architecture used in Stochastic Value Gradients."""
from dataclasses import dataclass, field

from dataclasses_json import DataClassJsonMixin
from gym.spaces import Box
from nnrl.nn.actor import StochasticActor
from nnrl.nn.critic import MLPVValue
from nnrl.nn.model import ResidualStochasticModel, SVGModel
from torch import nn

ModelSpec = SVGModel.spec_cls
CriticSpec = MLPVValue.spec_cls


@dataclass
class ActorSpec(StochasticActor.spec_cls):
    # pylint:disable=missing-class-docstring
    old_policy: bool = False


@dataclass
class SVGSpec(DataClassJsonMixin):
    # pylint:disable=missing-class-docstring
    model: ModelSpec = field(default_factory=ModelSpec)
    actor: ActorSpec = field(default_factory=ActorSpec)
    critic: CriticSpec = field(default_factory=CriticSpec)


class SVG(nn.Module):
    """Architecture used in Stochastic Value Gradients."""

    # pylint:disable=abstract-method
    spec_cls = SVGSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: SVGSpec):
        super().__init__()
        self._make_model(obs_space, action_space, spec.model)
        self._make_actor(obs_space, action_space, spec.actor)
        self._make_critic(obs_space, spec.critic)

    def _make_model(self, obs_space: Box, action_space: Box, spec: ModelSpec):
        model = SVGModel(obs_space, action_space, spec)
        self.model = ResidualStochasticModel(model) if spec.residual else model

    def _make_actor(self, obs_space: Box, action_space: Box, spec: ActorSpec):
        actor = StochasticActor(obs_space, action_space, spec)
        self.actor = actor.policy
        if spec.old_policy:
            self.old_actor = StochasticActor(obs_space, action_space, spec).policy

    def _make_critic(self, obs_space: Box, spec: CriticSpec):
        critic = MLPVValue(obs_space, spec)
        target_critic = MLPVValue(obs_space, spec)
        target_critic.load_state_dict(critic.state_dict())

        self.critic = critic
        self.target_critic = target_critic


class SoftSVG(SVG):
    """Architecture used for SVG in the Maximum Entropy framework."""

    def _make_actor(self, obs_space: Box, action_space: Box, spec: ActorSpec):
        actor = StochasticActor(obs_space, action_space, spec)
        self.actor = actor.policy
        self.alpha = actor.alpha



"""
From OpenAI Baselines:
https://github.com/openai/baselines/blob/master/baselines/ddpg/noise.py
"""
from dataclasses import dataclass, field

import numpy as np


@dataclass
class AdaptiveParamNoiseSpec:
    """Adaptive schedule for parameter noise exploration.

    Note that initial_stddev and curr_stddev refer to std of parameter noise,
    but desired_action_stddev refers to (as name suggests) the desired stddev
    in action space.
    """

    initial_stddev: float = 0.1
    desired_action_stddev: float = 0.2
    adaptation_coeff: float = 1.01
    curr_stddev: float = field(init=False)

    def __post_init__(self):
        self.curr_stddev = self.initial_stddev

    def adapt(self, distance):
        """Update current stddev based on action space distance."""
        if distance > self.desired_action_stddev:
            self.curr_stddev /= self.adaptation_coeff  # Decrease stddev.
        else:
            self.curr_stddev *= self.adaptation_coeff  # Increase stddev.


def ddpg_distance_metric(actions1, actions2):
    """Compute "distance" between actions taken by two policies at the same states.

    Expects numpy arrays.
    """
    diff = actions1 - actions2
    mean_diff = np.mean(np.square(diff), axis=0)
    dist = np.sqrt(np.mean(mean_diff))
    return dist



# pylint:disable=missing-module-docstring
import gym
import numpy as np
from gym.spaces import Box


class SinglePrecision(gym.Wrapper):
    """Ensures environment outputs are single-precision floats.

    Only compatible with continuous state-action environments.
    """

    def __init__(self, env: gym.Env):
        super().__init__(env)
        try:
            assert isinstance(env.observation_space, Box)
            assert isinstance(env.action_space, Box)
        except AssertionError as err:
            msg = f"{type(self).__name__} only compatible with Box obs/act spaces"
            raise ValueError(msg) from err

        self.observation_space = Box(
            low=env.observation_space.low.astype(np.float32),
            high=env.observation_space.high.astype(np.float32),
        )
        self.action_space = Box(
            low=env.action_space.low.astype(np.float32),
            high=env.action_space.high.astype(np.float32),
        )

    def reset(self, **kwargs):
        observation = self.env.reset(**kwargs)
        return self.observation(observation)

    def step(self, action):
        action = self.action(action)
        observation, reward, done, info = self.env.step(action)
        return self.observation(observation), reward, done, info

    def observation(self, observation: np.ndarray) -> np.ndarray:
        return observation.astype(self.observation_space.dtype, copy=False)

    def action(self, action: np.ndarray) -> np.ndarray:
        return action.astype(self.action_space.dtype, copy=False)



# pylint:disable=missing-module-docstring
from ray.rllib.utils import override
from ray.tune.logger import JsonLogger as _JsonLogger


class ProgressJsonLogger(_JsonLogger):
    """Custom JsonLogger which does not save the trial configuration every time."""

    @override(_JsonLogger)
    def on_result(self, result):
        tmp = result.copy()
        if "config" in tmp:
            del tmp["config"]
        super().on_result(tmp)



"""Loss functions for dual variables in maximum entropy RL."""
from typing import Callable, Tuple

import torch
from nnrl.nn.actor import StochasticPolicy
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from torch import Tensor

from raylab.utils.types import StatDict

from .abstract import Loss


class MaximumEntropyDual(Loss):
    """Loss function for the entropy coefficient in maximum entropy RL.

    Args:
        alpha: entropy coefficient
        actor: stochastic policy
        target_entropy: minimum entropy for policy
    """

    ENTROPY = "entropy"
    batch_keys = ("entropy", SampleBatch.CUR_OBS)

    def __init__(
        self,
        alpha: Callable[[], Tensor],
        actor: StochasticPolicy,
        target_entropy: float,
    ):
        self.alpha = alpha
        self.actor = actor
        self.target_entropy = target_entropy

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute entropy coefficient loss."""

        if self.ENTROPY in batch:
            entropy = batch[self.ENTROPY]
        else:
            with torch.no_grad():
                _, logp = self.actor(batch[SampleBatch.CUR_OBS])
                entropy = -logp

        alpha = self.alpha()
        entropy_diff = torch.mean(alpha * entropy - alpha * self.target_entropy)
        info = {
            "loss(alpha)": entropy_diff.item(),
            "curr_alpha": alpha.item(),
            "entropy": entropy.mean().item(),
        }
        return entropy_diff, info



"""Network and configurations for model-based DDPG algorithms."""
from dataclasses import dataclass, field

from gym.spaces import Box
from nnrl.nn.model import EnsembleSpec, build_ensemble

from .ddpg import DDPG, DDPGSpec


@dataclass
class ModelBasedDDPGSpec(DDPGSpec):
    """Specifications for model-based DDPG modules.

    Args:
        model: Specifications for stochastic dynamics model ensemble
        actor: Specifications for policy, behavior, and target policy
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides model, actor, and critic initializer
            specifications.
    """

    model: EnsembleSpec = field(default_factory=EnsembleSpec)

    def __post_init__(self):
        super().__post_init__()
        if self.initializer:
            self.model.initializer = self.initializer


class ModelBasedDDPG(DDPG):
    """NN module for Model-Based DDPG algorithms.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for model-based DDPG modules

    Attributes:
        model (SME): Stochastic dynamics model ensemble
        actor (DeterministicPolicy): The deterministic policy to be learned
        behavior (DeterministicPolicy): The policy for exploration
        target_actor (DeterministicPolicy): The policy used for estimating the
            arg max in Q-Learning
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = ModelBasedDDPGSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: ModelBasedDDPGSpec):
        super().__init__(obs_space, action_space, spec)

        self.models = build_ensemble(obs_space, action_space, spec.model)



"""Wrapper for introducing irrelevant state variables."""
import gym
import numpy as np
from gym.spaces import Box

from .mixins import IrrelevantRedundantMixin, RNGMixin
from .utils import assert_flat_box_space


class GaussianRandomWalks(IrrelevantRedundantMixin, RNGMixin, gym.ObservationWrapper):
    """Add gaussian random walk variables to the observations.

    Arguments:
        env: a gym environment instance
        size: the number of random walks to append to the observation.
        loc: mean of the Gaussian distribution
        scale: stddev of the Gaussian distribution
    """

    def __init__(self, env: gym.Env, size: int, loc: float = 0.0, scale: float = 1.0):
        assert_flat_box_space(env.observation_space, self)
        super().__init__(env)
        self._size = size
        self._loc = loc
        self._scale = scale
        self._random_walk = None

        original = self.env.observation_space
        low = np.concatenate([original.low, [-np.inf] * size])
        high = np.concatenate([original.high, [np.inf] * size])
        self.observation_space = Box(low=low, high=high, dtype=original.dtype)

        self._set_reward_if_possible()
        self._set_termination_if_possible()

    @property
    def added_size(self):
        return self._size

    def _added_vars(self, observation: np.ndarray) -> np.ndarray:
        self._random_walk = self._random_walk + self.np_random.normal(
            loc=self._loc, scale=self._scale, size=self._size
        )
        return self._random_walk

    def reset(self, **kwargs):
        self._random_walk = self.np_random.normal(
            loc=self._loc, scale=self._scale, size=self._size
        )
        return super().reset(**kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raylab',
 'raylab.agents',
 'raylab.agents.acktr',
 'raylab.agents.mage',
 'raylab.agents.mbpo',
 'raylab.agents.naf',
 'raylab.agents.sac',
 'raylab.agents.sop',
 'raylab.agents.svg',
 'raylab.agents.svg.inf',
 'raylab.agents.svg.one',
 'raylab.agents.svg.soft',
 'raylab.agents.td3',
 'raylab.agents.trpo',
 'raylab.cli',
 'raylab.envs',
 'raylab.envs.environments',
 'raylab.envs.wrappers',
 'raylab.execution',
 'raylab.logger',
 'raylab.policy',
 'raylab.policy.losses',
 'raylab.policy.model_based',
 'raylab.policy.modules',
 'raylab.tune',
 'raylab.utils',
 'raylab.utils.exploration']

package_data = \
{'': ['*']}

install_requires = \
['bokeh<2.3.4',
 'click>=7.1.2,<8.0.0',
 'dataclasses-json>=0.5.1,<0.6.0',
 'dm-tree>=0.1.5,<0.2.0',
 'nnrl>=0.1.0,<0.2.0',
 'opencv-contrib-python>=4.4.0,<5.0.0',
 'opencv-python>=4.2.0,<5.0.0',
 'poetry-version>=0.1.5,<0.2.0',
 'pytorch-lightning>=1.3.8,<1.6.0',
 'ray[rllib,tune]<1.7',
 'sklearn>=0.0,<0.1',
 'streamlit>=0.62,<0.86',
 'tabulate>=0.8.7,<0.9.0',
 'torch>=1.5.1,<2.0.0']

extras_require = \
{':python_version >= "3.7" and python_version < "4.0"': ['cachetools>=4.1.0,<5.0.0']}

entry_points = \
{'console_scripts': ['raylab = raylab.cli:raylab']}

setup_kwargs = {
    'name': 'raylab',
    'version': '0.16.1',
    'description': 'Reinforcement learning algorithms in RLlib and PyTorch.',
    'long_description': '======\nraylab\n======\n\n|PyPI| |Tests| |Dependabot| |License| |CodeStyle|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/raylab?logo=PyPi&logoColor=white&color=blue\n      :alt: PyPI\n\n.. |Tests| image:: https://img.shields.io/github/workflow/status/angelolovatto/raylab/Poetry%20package?label=tests&logo=GitHub\n       :alt: GitHub Workflow Status\n\n.. |Dependabot| image:: https://api.dependabot.com/badges/status?host=github&repo=angelolovatto/raylab\n        :target: https://dependabot.com\n\n.. |License| image:: https://img.shields.io/github/license/angelolovatto/raylab?color=blueviolet&logo=github\n         :alt: GitHub\n\n.. |CodeStyle| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n           :target: https://github.com/psf/black\n\n\nReinforcement learning algorithms in `RLlib <https://github.com/ray-project/ray/tree/master/rllib>`_\nand `PyTorch <https://pytorch.org>`_.\n\n\nInstallation\n------------\n\n.. code:: bash\n\n          pip install raylab\n\n\nQuickstart\n----------\n\nRaylab provides agents and environments to be used with a normal RLlib/Tune setup.\nYou can an agent\'s name (from the `Algorithms`_ section) to :code:`raylab info list` to list its top-level configurations:\n\n.. code-block:: zsh\n\n    raylab info list SoftAC\n\n.. code-block::\n\n    learning_starts: 0\n        Hold this number of timesteps before first training operation.\n    policy: {}\n        Sub-configurations for the policy class.\n    wandb: {}\n        Configs for integration with Weights & Biases.\n\n        Accepts arbitrary keyword arguments to pass to `wandb.init`.\n        The defaults for `wandb.init` are:\n        * name: `_name` property of the trainer.\n        * config: full `config` attribute of the trainer\n        * config_exclude_keys: `wandb` and `callbacks` configs\n        * reinit: True\n\n        Don\'t forget to:\n          * install `wandb` via pip\n          * login to W&B with the appropriate API key for your\n            team/project.\n          * set the `wandb/project` name in the config dict\n\n        Check out the Quickstart for more information:\n        `https://docs.wandb.com/quickstart`\n\nYou can add the :code:`--rllib` flag to get the descriptions for all the options common to RLlib agents\n(or :code:`Trainer`\\s)\n\nLaunching experiments can be done via the command line using :code:`raylab experiment` passing a file path\nwith an agent\'s configuration through the :code:`--config` flag.\nThe following command uses the cartpole `example <examples/PG/cartpole_defaults.py>`_ configuration file\nto launch an experiment using the vanilla Policy Gradient agent from the RLlib library.\n\n.. code-block:: zsh\n\n    raylab experiment PG --name PG -s training_iteration 10 --config examples/PG/cartpole_defaults.py\n\nYou can also launch an experiment from a Python script normally using Ray and Tune.\nThe following shows how you may use Raylab to perform an experiment comparing different\ntypes of exploration for the NAF agent.\n\n.. code-block:: python\n\n             import ray\n             from ray import tune\n             import raylab\n\n             def main():\n                 raylab.register_all_agents()\n                 raylab.register_all_environments()\n                 ray.init()\n                 tune.run(\n                     "NAF",\n                     local_dir="data/NAF",\n                     stop={"timesteps_total": 100000},\n                     config={\n                         "env": "CartPoleSwingUp-v0",\n                         "exploration_config": {\n                             "type": tune.grid_search([\n                                 "raylab.utils.exploration.GaussianNoise",\n                                 "raylab.utils.exploration.ParameterNoise"\n                             ])\n                         }\n                     },\n                     num_samples=10,\n                 )\n\n             if __name__ == "__main__":\n                 main()\n\n\nOne can then visualize the results using :code:`raylab dashboard`, passing the :code:`local_dir` used in the\nexperiment. The dashboard lets you filter and group results in a quick way.\n\n.. code-block:: zsh\n\n    raylab dashboard data/NAF/\n\n\n.. image:: https://i.imgur.com/bVc6WC5.png\n        :align: center\n\n\nYou can find the best checkpoint according to a metric (:code:`episode_reward_mean` by default)\nusing :code:`raylab find-best`.\n\n.. code-block:: zsh\n\n    raylab find-best data/NAF/\n\nFinally, you can pass a checkpoint to :code:`raylab rollout` to see the returns collected by the agent and\nrender it if the environment supports a visual :code:`render()` method. For example, you\ncan use the output of the :code:`find-best` command to see the best agent in action.\n\n\n.. code-block:: zsh\n\n    raylab rollout $(raylab find-best data/NAF/) --agent NAF\n\n\nAlgorithms\n----------\n\n+--------------------------------------------------------+-------------------------+\n| Paper                                                  | Agent Name              |\n+--------------------------------------------------------+-------------------------+\n| `Actor Critic using Kronecker-factored Trust Region`_  | ACKTR                   |\n+--------------------------------------------------------+-------------------------+\n| `Trust Region Policy Optimization`_                    | TRPO                    |\n+--------------------------------------------------------+-------------------------+\n| `Normalized Advantage Function`_                       | NAF                     |\n+--------------------------------------------------------+-------------------------+\n| `Stochastic Value Gradients`_                          | SVG(inf)/SVG(1)/SoftSVG |\n+--------------------------------------------------------+-------------------------+\n| `Soft Actor-Critic`_                                   | SoftAC                  |\n+--------------------------------------------------------+-------------------------+\n| `Streamlined Off-Policy`_ (DDPG)                       | SOP                     |\n+--------------------------------------------------------+-------------------------+\n| `Model-Based Policy Optimization`_                     | MBPO                    |\n+--------------------------------------------------------+-------------------------+\n| `Model-based Action-Gradient-Estimator`_               | MAGE                    |\n+--------------------------------------------------------+-------------------------+\n\n\n.. _`Actor Critic using Kronecker-factored Trust Region`: https://arxiv.org/abs/1708.05144\n.. _`Trust Region Policy Optimization`: http://proceedings.mlr.press/v37/schulman15.html\n.. _`Normalized Advantage Function`: http://proceedings.mlr.press/v48/gu16.html\n.. _`Stochastic Value Gradients`: http://papers.nips.cc/paper/5796-learning-continuous-control-policies-by-stochastic-value-gradients\n.. _`Soft Actor-Critic`: http://proceedings.mlr.press/v80/haarnoja18b.html\n.. _`Model-Based Policy Optimization`: http://arxiv.org/abs/1906.08253\n.. _`Streamlined Off-Policy`: https://arxiv.org/abs/1910.02208\n.. _`Model-based Action-Gradient-Estimator`: https://arxiv.org/abs/2004.14309\n\n\nCommand-line interface\n----------------------\n\n.. role:: bash(code)\n   :language: bash\n\nFor a high-level description of the available utilities, run :bash:`raylab --help`\n\n.. code:: bash\n\n    Usage: raylab [OPTIONS] COMMAND [ARGS]...\n\n      RayLab: Reinforcement learning algorithms in RLlib.\n\n    Options:\n      --help  Show this message and exit.\n\n    Commands:\n      dashboard    Launch the experiment dashboard to monitor training progress.\n      episodes     Launch the episode dashboard to monitor state and action...\n      experiment   Launch a Tune experiment from a config file.\n      find-best    Find the best experiment checkpoint as measured by a metric.\n      info         View information about an agent\'s config parameters.\n      rollout      Wrap `rllib rollout` with customized options.\n      test-module  Launch dashboard to test generative models from a checkpoint.\n\n\nPackages\n--------\n\nThe project is structured as follows\n::\n\n    raylab\n    |-- agents            # Trainer and Policy classes\n    |-- cli               # Command line utilities\n    |-- envs              # Gym environment registry and utilities\n    |-- logger            # Tune loggers\n    |-- policy            # Extensions and customizations of RLlib\'s policy API\n    |   |-- losses        # RL loss functions\n    |   |-- modules       # PyTorch neural network modules for TorchPolicy\n    |-- pytorch           # PyTorch extensions\n    |-- utils             # miscellaneous utilities\n',
    'author': 'Ângelo Gregório Lovatto',
    'author_email': 'angelolovatto@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/angelolovatto/raylab',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Losses for Stochastic Value Gradients."""
from typing import Callable, List, Tuple

import torch
from nnrl.nn.actor import Alpha, StochasticPolicy
from nnrl.nn.critic import VValue
from nnrl.nn.model import StochasticModel
from nnrl.types import TensorDict
from ray.rllib import SampleBatch
from ray.rllib.utils import override
from torch import Tensor, nn

from raylab.utils.types import RewardFn, StatDict

from .abstract import Loss
from .mixins import EnvFunctionsMixin

StateRepr = Callable[[Tensor, Tensor, Tensor], Tuple[Tensor, Tensor]]
ActionRepr = Callable[[Tensor, Tensor], Tuple[Tensor, Tensor]]


class OneStepSVG(EnvFunctionsMixin, Loss):
    """Loss function for Stochastic Value Gradients with 1-step bootstrapping.

    Args:
        model: Stochastic model that reproduces state and its log density
        actor: Stochastic policy that reproduces action and its log density
        critic: State-value function

    Attributes:
        gamma: discount factor
    """

    IS_RATIOS = "is_ratios"
    gamma: float = 0.99
    batch_keys: Tuple[str, str, str, str, str]

    def __init__(self, model: StochasticModel, actor: StochasticPolicy, critic: VValue):
        super().__init__()
        self.model = model
        self.actor = actor
        self.critic = critic

        self.batch_keys = (
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            SampleBatch.NEXT_OBS,
            SampleBatch.DONES,
            self.IS_RATIOS,
        )

    def __call__(self, batch: TensorDict) -> Tuple[Tensor, StatDict]:
        """Compute bootstrapped Stochatic Value Gradient loss."""
        assert self._env.reward is not None, "Reward function must be set!"
        obs, actions, next_obs, dones, is_ratios = self.unpack_batch(batch)

        state_val = self.one_step_reproduced_state_value(obs, actions, next_obs, dones)
        svg_loss = -torch.mean(is_ratios * state_val)
        return svg_loss, {"loss(actor)": svg_loss.item()}

    def one_step_reproduced_state_value(
        self, obs: Tensor, actions: Tensor, next_obs: Tensor, dones: Tensor
    ) -> Tensor:
        """Compute 1-step approximation of the state value on real transition."""
        _acts, _ = self.actor.reproduce(obs, actions)
        _next_obs, _ = self.model.reproduce(next_obs, self.model(obs, _acts))
        _rewards = self._env.reward(obs, _acts, _next_obs)
        _next_vals = self.critic(_next_obs)
        return torch.where(dones, _rewards, _rewards + self.gamma * _next_vals)


class OneStepSoftSVG(OneStepSVG):
    """Loss function for bootstrapped maximum entropy Stochastic Value Gradients.

    Args:
        model: Stochastic model that reproduces state and its log density
        actor: Stochastic policy that reproduces action and its log density
        critic: State-value function
        alpha: Entropy coefficient

    Attributes:
        gamma: Discount factor
    """

    gamma: float = 0.99

    def __init__(
        self,
        model: StochasticModel,
        actor: StochasticPolicy,
        critic: VValue,
        alpha: Alpha,
    ):
        self.alpha = alpha
        super().__init__(model, actor, critic)

    @override(OneStepSVG)
    def one_step_reproduced_state_value(
        self, obs: Tensor, actions: Tensor, next_obs: Tensor, dones: Tensor
    ) -> Tensor:
        _actions, _logp = self.actor.reproduce(obs, actions)
        _next_obs, _ = self.model.reproduce(next_obs, self.model(obs, _actions))
        _rewards = self._env.reward(obs, _actions, _next_obs)
        _entropy = -_logp
        _augmented_rewards = _rewards + _entropy * self.alpha()
        _next_vals = self.critic(_next_obs)

        return torch.where(
            dones,
            _augmented_rewards,
            _augmented_rewards + self.gamma * _next_vals,
        )


class TrajectorySVG(EnvFunctionsMixin, Loss):
    """Loss function for Stochastic Value Gradients on full trajectory.

    Args:
        model: model that reproduces state and its log density
        actor: policy that reproduces action and its log density
        critic: state-value function
    """

    batch_keys: Tuple[str, str, str] = (
        SampleBatch.CUR_OBS,
        SampleBatch.ACTIONS,
        SampleBatch.NEXT_OBS,
    )

    def __init__(self, model: StochasticModel, actor: StochasticPolicy, critic: VValue):
        super().__init__()
        self.model = model
        self.actor = actor
        self.critic = critic

        self._rollout = None

    def set_reward_fn(self, *args, **kwargs):
        super().set_reward_fn(*args, **kwargs)
        self._rollout = ReproduceRewards(self.actor, self.model, self._env.reward)

    def compile(self):
        """Compile the rollout module to TorchScript."""
        self._rollout = torch.jit.script(self._rollout)

    def __call__(self, episodes: List[TensorDict]) -> Tuple[Tensor, StatDict]:
        """Compute Stochatic Value Gradient loss given full trajectories."""
        assert (
            self._rollout is not None
        ), "Rollout module not set. Did you call `set_reward_fn`?"

        total_ret = 0
        for episode in episodes:
            init_obs = episode[SampleBatch.CUR_OBS][0]
            actions = episode[SampleBatch.ACTIONS]
            next_obs = episode[SampleBatch.NEXT_OBS]

            _, _, rewards = self._rollout(actions, next_obs, init_obs)
            total_ret += rewards.sum()

        sim_return_mean = total_ret / len(episodes)
        loss = -sim_return_mean
        info = {"loss(actor)": loss.item(), "sim_return_mean": sim_return_mean.item()}
        return loss, info


class ReproduceRewards(nn.Module):
    """Unrolls a policy, model and reward function given a trajectory.

    Args:
        model: transition model that reproduces state and its log density
        actor: stochastic policy that reproduces action and its log density
        reward_fn: reward function for state, action, and next state tuples
    """

    # pylint:disable=abstract-method
    def __init__(
        self, policy: StochasticPolicy, model: StochasticModel, reward_fn: RewardFn
    ):
        super().__init__()
        self.policy = policy
        self.model = model
        self.reward_fn = reward_fn

    @override(nn.Module)
    def forward(
        self, acts: Tensor, next_obs: Tensor, init_ob: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """Reproduce a sequence of actions, obsevations, and rewards.

        Args:
            acts: the sequence of actions
            next_obs: the sequence of observations that resulted from the
                application of the previous actions
            init_ob: the initial observation

        Returns:
            The observations, actions and rewards from the reproduced trajectory

        Note:
            Assumes the first tensor dimension of `acts` and `next_obs`
            corresponds to the timestep and iterates over it.
        """
        # pylint:disable=arguments-differ
        obs_seq = []
        act_seq = []
        reward_seq = []
        cur_ob = init_ob
        for act, next_ob in zip(acts, next_obs):
            _act, _ = self.policy.reproduce(cur_ob, act)
            act_seq += [_act]
            _next_ob, _ = self.model.reproduce(next_ob, self.model(cur_ob, _act))
            obs_seq += [_next_ob]
            reward_seq += [self.reward_fn(cur_ob, _act, _next_ob)]
            cur_ob = _next_ob
        return torch.stack(obs_seq), torch.stack(act_seq), torch.stack(reward_seq)



"""CLI for finding the best checkpoint of an experiment."""
import click

from .utils import initialize_raylab


def get_last_checkpoint_path(logdir):
    """Retrieve the path of the last checkpoint given a Trial logdir."""
    import os.path as osp
    from glob import glob

    last_checkpoint_basedir = sorted(
        glob(osp.join(logdir, "checkpoint_*")), key=lambda p: p.split("_")[-1]
    )[-1]
    last_checkpoint_path = osp.join(
        last_checkpoint_basedir, osp.basename(last_checkpoint_basedir).replace("_", "-")
    )
    return last_checkpoint_path


@click.command()
@click.argument(
    "logdir",
    nargs=1,
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
)
@click.option(
    "--metric",
    default="episode_reward_mean",
    show_default=True,
    help="Key for trial info to order on.",
)
@click.option(
    "--mode",
    type=click.Choice("max min".split()),
    default="max",
    show_default=True,
    help="Criterion to order trials by.",
)
@initialize_raylab
def find_best(logdir, metric, mode):
    """Find the best experiment checkpoint as measured by a metric."""
    import logging

    from ray.tune.analysis import Analysis

    logging.getLogger("ray.tune").setLevel("ERROR")

    analysis = Analysis(logdir)
    best_logdir = analysis.get_best_logdir(metric, mode=mode)
    last_checkpoint_path = get_last_checkpoint_path(best_logdir)
    click.echo(last_checkpoint_path)



# pylint:disable=missing-module-docstring
from typing import Tuple

import torch
from nnrl.nn.actor import DeterministicPolicy
from torch import Tensor

from raylab.policy.action_dist import BaseActionDist

from .base import Model
from .random_uniform import RandomUniform


class GaussianNoise(RandomUniform):
    """Adds fixed additive gaussian exploration noise to actions.

    Args:
        noise_stddev: Standard deviation of the Gaussian samples
    """

    valid_behavior_cls = DeterministicPolicy

    def __init__(self, *args, noise_stddev: float = 0.3, **kwargs):
        super().__init__(*args, **kwargs)
        self._noise_stddev = noise_stddev

    def get_exploration_action(
        self,
        *,
        action_distribution: BaseActionDist,
        timestep: int,
        explore: bool = True,
    ):
        if explore:
            if timestep < self._pure_exploration_steps:
                return super().get_exploration_action(
                    action_distribution=action_distribution,
                    timestep=timestep,
                    explore=explore,
                )
            return self._inject_gaussian_noise(action_distribution)
        return action_distribution.deterministic_sample()

    def _inject_gaussian_noise(
        self, action_distribution: BaseActionDist
    ) -> Tuple[Tensor, None]:
        module, inputs = action_distribution.module, action_distribution.inputs
        unconstrained_action = module.behavior.unconstrained_action(**inputs)
        unconstrained_action += (
            torch.randn_like(unconstrained_action) * self._noise_stddev
        )
        action = module.behavior.squash_action(unconstrained_action)
        return action, None

    @classmethod
    def check_model_compat(cls, model: Model):
        RandomUniform.check_model_compat(model)
        assert model is not None, f"{cls} exploration needs access to the NN."
        assert hasattr(
            model, "behavior"
        ), f"NN model {type(model)} has no behavior attribute."
        assert isinstance(model.behavior, cls.valid_behavior_cls), (
            f"Expected behavior to be an instance of {cls.valid_behavior_cls};"
            " found {type(model.behavior)} instead."
        )



"""Custom Gym wrappers for environments."""
from .correlated_irrelevant import CorrelatedIrrelevant
from .gaussian_random_walks import GaussianRandomWalks
from .linear_redundant import LinearRedundant
from .nonlinear_redundant import NonlinearRedundant
from .random_irrelevant import RandomIrrelevant
from .single_precision import SinglePrecision
from .time_aware_env import AddRelativeTimestep

__all__ = [
    "AddRelativeTimestep",
    "CorrelatedIrrelevant",
    "GaussianRandomWalks",
    "LinearRedundant",
    "NonlinearRedundant",
    "RandomIrrelevant",
    "SinglePrecision",
]



# pylint:disable=missing-module-docstring
from ray.rllib.utils import override

from .random_uniform import RandomUniform


class StochasticActor(RandomUniform):
    """Exploration class compatible with StochasticActorMixin submodules."""

    @override(RandomUniform)
    def get_exploration_action(self, *, action_distribution, timestep, explore=True):
        # pylint:disable=too-many-arguments
        if explore:
            if timestep < self._pure_exploration_steps:
                return super().get_exploration_action(
                    action_distribution=action_distribution,
                    timestep=timestep,
                    explore=explore,
                )
            return action_distribution.sample()
        return action_distribution.deterministic_sample()



"""Utilities for loading and processing experiment results."""
# pylint:disable=missing-docstring,unsubscriptable-object
from __future__ import annotations

import itertools
import json
import logging
import os
from ast import literal_eval
from collections import namedtuple
from functools import reduce
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)

_NUMERIC_KINDS = set("uifc")
ExperimentData = namedtuple("ExperimentData", ["progress", "params", "flat_params"])
Folder = namedtuple("Folder", ["path", "files"])

###############################################################################
# Experiment data loading from disk
###############################################################################


def load_exps_data(
    directories: Union[list[str], str],
    progress_prefix: str = "progress",
    config_prefix: str = "params",
    error_prefix: str = "error",
    include_errors: bool = False,
    verbose: bool = False,
) -> list[ExperimentData]:
    """Extract experiment output from directories.

    Recursively searchs directories for experiment folders and loads data
    (csv logs, hyperparameters). Returns each experiment output as an
    ExperimentData object.

    Args:
        directories: directories to recursively search for experiment data
        progress_prefix: prefix for progress files containing logs for each run
        config_prefix: prefix for config files containing hyperparameters for
            each run
        error_prefix: prefix for files containing console logs for crashed runs
        include_errors: whether to include experiments that have at least one
            error file, identified by the `error_prefix` parameter, indicating
            that the run crashed
        verbose: whether to log intermediary steps from this function to stdout

    Returns:
        A list of ExperimentData containing each experiment's output
    """
    # pylint:disable=too-many-arguments
    if isinstance(directories, str):
        directories = [directories]

    def isprogress(file: str) -> bool:
        return file.startswith(progress_prefix) and file.endswith(".csv")

    def isconfig(file: str) -> bool:
        return file.startswith(config_prefix) and file.endswith(".json")

    exp_folders = get_folders_with_target_files(directories, isprogress)
    if not include_errors:
        exp_folders = [
            d
            for d in exp_folders
            if not any(f.startswith(error_prefix) for f in d.files)
        ]
    if verbose:
        logger.info("finished walking exp folders")

    exps_data = read_exp_folder_data(exp_folders, isprogress, isconfig, verbose=verbose)
    return exps_data


# noinspection PyTypeChecker,PyUnresolvedReferences
def get_folders_with_target_files(
    directories: list[str], istarget: callable[[str], bool]
):
    """Return folders that havy any file of interest."""
    # (path, subpath, files) for each dir
    paths_per_dir = map(os.walk, directories)
    # (path, subpath, files) for all dirs
    all_paths = reduce(itertools.chain, paths_per_dir, ())
    # only (path, files)
    paths_and_files = map(lambda x: (x[0], x[2]), all_paths)
    # entries that have >0 files (may be unnecessary)
    paths_with_files = filter(lambda x: x[1], paths_and_files)
    # entries that have target files
    paths_with_target_files = filter(
        lambda x: any(istarget(y) for y in x[1]), paths_with_files
    )
    # Folders with at least one target file
    folders = map(lambda x: Folder(*x), paths_with_target_files)
    return list(folders)


# noinspection PyUnresolvedReferences
def read_exp_folder_data(
    exp_folders: list[Folder],
    isprogress: callable[[str], bool],
    isconfig: callable[[str], bool],
    verbose: bool = False,
) -> list[ExperimentData]:
    exps_data = []
    for folder in exp_folders:
        try:
            progress = load_progress(
                os.path.join(folder.path, first_that(isprogress, folder.files)),
                verbose=verbose,
            )
            params = load_params(
                os.path.join(folder.path, first_that(isconfig, folder.files))
            )
            if "trial_id" in progress:
                params["id"] = progress["trial_id"][0]
            exps_data.append(
                ExperimentData(
                    progress=progress, params=params, flat_params=flatten_dict(params)
                )
            )
        except (IOError, pd.errors.EmptyDataError) as error:
            if verbose:
                logger.error(error)

    return exps_data


# noinspection PyUnresolvedReferences
def first_that(criterion: callable[[Any], bool], lis: list[Any]) -> Any:
    return next((x for x in lis if criterion(x)), None)


def load_progress(progress_path: str, verbose: bool = False) -> pd.DataFrame:
    if verbose:
        # pylint:disable=logging-format-interpolation
        logger.info("Reading {}".format(progress_path))

    if progress_path.endswith(".csv"):
        return pd.read_csv(progress_path, index_col=False, comment="#")

    # Assume progress is in JSON format
    dicts = []
    with open(progress_path, "rt") as file:
        for line in file:
            dicts.append(json.loads(line))
    return pd.DataFrame(dicts)


def flatten_dict(dic: dict) -> dict:
    flat_params = dict()
    for key, val in dic.items():
        if isinstance(val, dict):
            # pylint:disable=fixme
            # FIXME: calling flatten_dict twice seems to be a mistake
            val = flatten_dict(val)
            for subk, subv in flatten_dict(val).items():
                flat_params[key + "/" + subk] = subv
        else:
            flat_params[key] = val
    return flat_params


def load_params(params_json_path: Optional[str]) -> dict:
    if params_json_path is None:
        return dict(exp_name="experiment")

    with open(params_json_path, "r") as file:
        data = json.load(file)
        if "args_data" in data:
            del data["args_data"]
        if "exp_name" not in data:
            data["exp_name"] = params_json_path.split("/")[-3]
    return data


###############################################################################
# Experiment data filtering in memory
###############################################################################


class Selector:
    # noinspection PyUnresolvedReferences
    def __init__(
        self,
        exps_data: list[ExperimentData],
        filters: Optional[tuple[callable[[ExperimentData], bool]]] = None,
    ):
        self._exps_data = exps_data
        self._filters = tuple() if filters is None else tuple(filters)

    def where(self, key: str, val: str) -> Selector:
        return Selector(
            self._exps_data,
            self._filters
            + (lambda exp: str(exp.flat_params.get(key, None)) == str(val),),
        )

    def where_not(self, key: str, val: str) -> Selector:
        return Selector(
            self._exps_data,
            self._filters
            + (lambda exp: str(exp.flat_params.get(key, None)) != str(val),),
        )

    def _check_exp(self, exp: ExperimentData) -> bool:
        return all(condition(exp) for condition in self._filters)

    def extract(self) -> list[ExperimentData]:
        return list(filter(self._check_exp, self._exps_data))


def filter_and_split_experiments(
    exps_data: list[ExperimentData],
    split: Optional[str] = None,
    include: tuple[str, ...] = (),
    exclude: tuple[str, ...] = (),
) -> tuple[list[Selector], list[str]]:
    selector = Selector(exps_data)
    for key, val in include:
        selector = selector.where(key, val)
    for key, val in exclude:
        selector = selector.where_not(key, val)

    if split is not None:
        exps_data = selector.extract()
        values = dict(sorted(extract_distinct_params(exps_data))).get(split, [])
        titles = list(map(str, values))
        selectors = [selector.where(split, t) for t in titles]
    else:
        selectors = [selector]
        titles = ["Experiment"]

    return selectors, titles


def extract_distinct_params(
    exps_data: list[ExperimentData],
    excluded_params: tuple[str, ...] = ("seed", "log_dir"),
) -> list[tuple[str, Any]]:
    repr_config_pairs = [repr(kv) for d in exps_data for kv in d.flat_params.items()]
    uniq_pairs = list(set(repr_config_pairs))
    evald_pairs = map(literal_eval, uniq_pairs)
    stringified_pairs = sorted(
        evald_pairs, key=lambda x: tuple("" if it is None else str(it) for it in x)
    )

    proposals = [
        (k, [x[1] for x in v])
        for k, v in itertools.groupby(stringified_pairs, lambda x: x[0])
    ]

    filtered = [
        (k, v)
        for (k, v) in proposals
        if v
        and all(k != excluded_param for excluded_param in excluded_params)
        and len(v) > 1
    ]
    return filtered


def is_numeric(array) -> bool:
    """Determine whether the argument has a numeric datatype, when
    converted to a NumPy array.

    Booleans, unsigned integers, signed integers, floats and complex
    numbers are the kinds of numeric datatype.

    Parameters
    ----------
    array : array-like
        The array to check.

    Returns
    -------
    is_numeric : `bool`
        True if the array has a numeric datatype, False if not.

    """
    return np.asarray(array).dtype.kind in _NUMERIC_KINDS


def is_increasing_key(key: str, exps_data: list[ExperimentData]) -> bool:
    for exp in exps_data:
        if key in exp.progress and not is_increasing(exp.progress[key]):
            return False
    return True


def is_increasing(arr: np.ndarray) -> bool:
    arr = np.asarray(arr)
    if not is_numeric(arr):
        return False

    arr = arr[~np.isnan(arr)]
    if len(arr) == 0:
        return False

    return np.all(np.less_equal(arr[:-1], arr[1:])) and np.max(arr) >= np.min(arr)


def get_plottable_keys(exps_data: list[ExperimentData]) -> list[str]:
    return sorted(
        list(
            set(
                col
                for exp in exps_data
                for col in exp.progress.columns.to_list()
                if is_numeric(exp.progress[col])
            )
        )
    )


def get_x_plottable_keys(
    plottable_keys: list[str], exps_data: list[ExperimentData]
) -> list[str]:
    return [key for key in plottable_keys if is_increasing_key(key, exps_data)]



"""Mixins for loss functions."""
from dataclasses import dataclass
from typing import Optional, Tuple

import numpy as np
from torch import nn

from raylab.utils.types import RewardFn, TerminationFn


@dataclass
class EnvFunctions:
    """Collection of environment emulating functions."""

    reward: Optional[RewardFn] = None
    termination: Optional[TerminationFn] = None

    @property
    def initialized(self):
        """Whether or not all functions are set."""
        return self.reward is not None and self.termination is not None


class EnvFunctionsMixin:
    """Adds private, externally set environment functions.

    The resulting loss function will have an `_env` private attribute. The user
    can use the `set_reward_fn` and `set_termination_fn` methods to set the
    environment functions. When both are set, `_env.initialized` will be True.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._env = EnvFunctions()

    def set_reward_fn(self, function: RewardFn):
        """Set reward function to provided callable."""
        self._env.reward = function

    def set_termination_fn(self, function: TerminationFn):
        """Set termination function to provided callable."""
        self._env.termination = function

    def check_env_fns(self):
        """Assert env functions have been set.

        Raises:
            AssertionError: If reward or termination functions were not set
        """
        assert self._env.initialized, "Reward or termination functions missing."


class UniformModelPriorMixin:
    """Add methods for using model ensembles with uniform prior distribution.

    Expects a model ensemble as a `models` instance attribute.

    Attributes:
        models: Module list of dynamics models
    """

    models: nn.ModuleList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rng = np.random.default_rng()

    def seed(self, seed: int):
        """Seeds the RNG for choosing a model from the ensemble."""
        self._rng = np.random.default_rng(seed)

    def sample_model(self) -> Tuple[nn.Module, int]:
        """Return a model and its index sampled uniformly at random."""
        models = self.models
        idx = self._rng.integers(len(models))
        return models[idx], idx



# pylint:disable=missing-module-docstring
import contextlib
import functools
import logging
import os
import warnings


@contextlib.contextmanager
def lightning_warnings_only():
    logger = logging.getLogger("lightning")
    level = logger.getEffectiveLevel()
    try:
        logger.setLevel(logging.WARNING)
        yield
    finally:
        logger.setLevel(level)


@contextlib.contextmanager
def suppress_dataloader_warnings():
    """Ignore PyTorch Lightning warnings regarding num of dataloader workers."""
    ignore = functools.partial(
        warnings.filterwarnings,
        "ignore",
        module="pytorch_lightning.trainer.data_loading",
    )
    with warnings.catch_warnings():
        ignore(message="The dataloader, .+, does not have many workers")
        ignore(message="One of given dataloaders is None")
        yield


def context_to_devnull(func, *, context_fn):
    """Decorator for applying a file-dependent context manager with devnull."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with open(os.devnull, "w") as devnull:
            with context_fn(devnull):
                return func(*args, **kwargs)

    return wrapper


suppress_stdout = functools.partial(
    context_to_devnull, context_fn=contextlib.redirect_stdout
)
suppress_stderr = functools.partial(
    context_to_devnull, context_fn=contextlib.redirect_stderr
)



"""TRPO policy implemented in PyTorch."""
import numpy as np
import torch
from nnrl.optim import build_optimizer
from nnrl.optim.hessian_free import (
    conjugate_gradient,
    hessian_vector_product,
    line_search,
)
from nnrl.utils import flat_grad
from ray.rllib import SampleBatch
from ray.rllib.evaluation.postprocessing import Postprocessing, compute_advantages
from ray.rllib.utils import override
from torch import nn
from torch.nn.utils import parameters_to_vector, vector_to_parameters

from raylab.options import configure, option
from raylab.policy import TorchPolicy, learner_stats
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.utils.dictionaries import get_keys
from raylab.utils.explained_variance import explained_variance

LINESEARCH_DEFAULTS = {
    "accept_ratio": 0.1,
    "backtrack_ratio": 0.8,
    "max_backtracks": 15,
    "atol": 1e-7,
}


@configure
@option("delta", 0.01, help="Trust region constraint")
@option(
    "fvp_samples",
    10,
    help="Number of actions to sample per state for Fisher vector-product calculation",
)
@option("lambda", 0.97, help=r"For GAE(\gamma, \lambda)")
@option("val_iters", 80, help="Number of iterations to fit value function")
@option("use_gae", True, help="Whether to use Generalized Advantage Estimation")
@option("cg_iters", 10, help="Number of iterations for Conjugate Gradient")
@option(
    "cg_damping",
    1e-3,
    help="Damping factor to avoid singular matrix multiplication in Conjugate Gradient",
)
@option(
    "line_search",
    True,
    help="""
    Whether to use a line search to calculate policy update.
    Effectively turns TRPO into Natural PG when turned off.
    """,
)
@option("line_search_options", LINESEARCH_DEFAULTS)
@option("module/type", "TRPO")
@option("optimizer", {"type": "Adam", "lr": 1e-3}, override=True)
@option("exploration_config/type", "raylab.utils.exploration.StochasticActor")
class TRPOTorchPolicy(TorchPolicy):
    """Policy class for Trust Region Policy Optimization."""

    # pylint:disable=abstract-method
    dist_class = WrapStochasticPolicy

    @override(TorchPolicy)
    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        optimizers.update(
            critic=build_optimizer(self.module.critic, self.config["optimizer"])
        )
        return optimizers

    @torch.no_grad()
    @override(TorchPolicy)
    def postprocess_trajectory(
        self, sample_batch, other_agent_batches=None, episode=None
    ):
        sample_batch = super().postprocess_trajectory(
            sample_batch, other_agent_batches=other_agent_batches, episode=episode
        )

        last_obs = self.convert_to_tensor(sample_batch[SampleBatch.NEXT_OBS][-1])
        last_r = self.module.critic(last_obs).squeeze(-1).numpy()

        cur_obs = self.convert_to_tensor(sample_batch[SampleBatch.CUR_OBS])
        sample_batch[SampleBatch.VF_PREDS] = (
            self.module.critic(cur_obs).squeeze(-1).numpy()
        )
        sample_batch = compute_advantages(
            sample_batch,
            last_r,
            gamma=self.config["gamma"],
            lambda_=self.config["lambda"],
            use_gae=self.config["use_gae"],
        )
        return sample_batch

    @learner_stats
    @override(TorchPolicy)
    def learn_on_batch(self, samples):
        batch_tensors = self.lazy_tensor_dict(samples)
        info = {}

        info.update(self._update_actor(batch_tensors))
        info.update(self._update_critic(batch_tensors))
        info.update(self.extra_grad_info(batch_tensors))
        info.update(self.get_exploration_info())

        return info

    def _update_actor(self, batch_tensors):
        info = {}
        cur_obs, actions, advantages = get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            Postprocessing.ADVANTAGES,
        )
        advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)

        # Compute Policy Gradient
        surr_loss = -(self.module.actor.log_prob(cur_obs, actions) * advantages).mean()
        pol_grad = flat_grad(surr_loss, self.module.actor.parameters())
        info["grad_norm(pg)"] = pol_grad.norm().item()

        # Compute Natural Gradient
        descent_step, cg_info = self._compute_descent_step(pol_grad, cur_obs)
        info["grad_norm(nat)"] = descent_step.norm().item()
        info.update(cg_info)

        # Perform Line Search
        if self.config["line_search"]:
            new_params, line_search_info = self._perform_line_search(
                pol_grad,
                descent_step,
                surr_loss,
                batch_tensors,
            )
            info.update(line_search_info)
        else:
            new_params = (
                parameters_to_vector(self.module.actor.parameters()) - descent_step
            )

        vector_to_parameters(new_params, self.module.actor.parameters())
        return info

    def _compute_descent_step(self, pol_grad, obs):
        """Approximately compute the Natural gradient using samples.

        This is based on the Fisher Matrix formulation as the hessian of the average
        entropy. For more information, see:
        https://en.wikipedia.org/wiki/Fisher_information#Matrix_form

        Args:
            pol_grad (Tensor): The vector to compute the Fisher vector product with.
            obs (Tensor): The observations to evaluate the policy in.
        """
        config = self.config
        params = list(self.module.actor.parameters())
        with torch.no_grad():
            ent_acts, _ = self.module.actor.sample(obs, (config["fvp_samples"],))

        def entropy():
            return self.module.actor.log_prob(obs, ent_acts).neg().mean()

        def fvp(vec):
            return hessian_vector_product(entropy(), params, vec)

        descent_direction, elapsed_iters, residual = conjugate_gradient(
            lambda x: fvp(x) + config["cg_damping"] * x,
            pol_grad,
            cg_iters=config["cg_iters"],
        )

        fisher_norm = pol_grad.dot(descent_direction)
        delta = config["delta"]
        scale = 0 if fisher_norm < 0 else torch.sqrt(2 * delta / (fisher_norm + 1e-8))

        descent_direction = descent_direction * scale
        return descent_direction, {"cg_iters": elapsed_iters, "cg_residual": residual}

    def _perform_line_search(self, pol_grad, descent_step, surr_loss, batch_tensors):
        expected_improvement = pol_grad.dot(descent_step).item()

        cur_obs, actions, old_logp, advantages = get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            SampleBatch.ACTION_LOGP,
            Postprocessing.ADVANTAGES,
        )

        @torch.no_grad()
        def f_barrier(params):
            vector_to_parameters(params, self.module.actor.parameters())
            new_logp = self.module.actor.log_prob(cur_obs, actions)
            surr_loss = self._compute_surr_loss(old_logp, new_logp, advantages)
            avg_kl = torch.mean(old_logp - new_logp)
            return surr_loss.item() if avg_kl < self.config["delta"] else np.inf

        new_params, expected_improvement, improvement = line_search(
            f_barrier,
            parameters_to_vector(self.module.actor.parameters()),
            descent_step,
            expected_improvement,
            y_0=surr_loss.item(),
            **self.config["line_search_options"],
        )
        improvement_ratio = (
            improvement / expected_improvement if expected_improvement else np.nan
        )
        info = {
            "expected_improvement": expected_improvement,
            "actual_improvement": improvement,
            "improvement_ratio": improvement_ratio,
        }
        return new_params, info

    @staticmethod
    def _compute_surr_loss(old_logp, new_logp, advantages):
        return -torch.mean(torch.exp(new_logp - old_logp) * advantages)

    def _update_critic(self, batch_tensors):
        info = {}
        mse = nn.MSELoss()

        cur_obs, value_targets, value_preds = get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            Postprocessing.VALUE_TARGETS,
            SampleBatch.VF_PREDS,
        )

        for _ in range(self.config["val_iters"]):
            with self.optimizers.optimize("critic"):
                loss = mse(self.module.critic(cur_obs).squeeze(-1), value_targets)
                loss.backward()

        info["vf_loss"] = loss.item()
        info["explained_variance"] = explained_variance(
            value_targets.numpy(), value_preds.numpy()
        )
        return info

    @torch.no_grad()
    def extra_grad_info(self, batch_tensors):  # pylint:disable=unused-argument
        """Return statistics right after components are updated."""
        cur_obs, actions, old_logp, value_targets, value_preds = get_keys(
            batch_tensors,
            SampleBatch.CUR_OBS,
            SampleBatch.ACTIONS,
            SampleBatch.ACTION_LOGP,
            Postprocessing.VALUE_TARGETS,
            SampleBatch.VF_PREDS,
        )

        info = {
            "kl_divergence": torch.mean(
                old_logp - self.module.actor.log_prob(cur_obs, actions)
            ).item(),
            "entropy": torch.mean(-old_logp).item(),
            "perplexity": torch.mean(-old_logp).exp().item(),
            "explained_variance": explained_variance(
                value_targets.numpy(), value_preds.numpy()
            ),
            "grad_norm(critic)": nn.utils.clip_grad_norm_(
                self.module.critic.parameters(), float("inf")
            ).item(),
        }
        return info



"""Trainer for TD3."""
from raylab.agents import Trainer
from raylab.agents.off_policy import OffPolicyMixin
from raylab.options import configure

from .policy import TD3TorchPolicy


@configure
@OffPolicyMixin.add_options
class TD3Trainer(OffPolicyMixin, Trainer):
    """Trainer for the Twin Delayed Deep Deterministic Policy Gradient algorithm."""

    _name = "TD3"
    _policy_class = TD3TorchPolicy



"""Registry of modules for PyTorch policies."""
from gym.spaces import Space
from torch import nn

from .ddpg import DDPG
from .mage import MAGE
from .mb_ddpg import ModelBasedDDPG
from .mb_sac import ModelBasedSAC
from .naf import NAF
from .sac import SAC
from .sop import SOP
from .svg import SVG, SoftSVG
from .td3 import TD3
from .trpo import TRPO

MODULES = {}


class RepeatedModuleNameError(Exception):
    """Exception raised for attempting to register a repeated module name.

    Args:
        cls: NN module class
    """

    def __init__(self, cls: type):
        super().__init__(f"Module class {cls.__name__} already in catalog")


class UnknownModuleError(Exception):
    """Exception raised for attempting to query an unkown module.

    Args:
        name: NN module name
    """

    def __init__(self, name: str):
        super().__init__(f"No module registered with name '{name}' in catalog")


def register(cls: type):
    """Register module class in catalog by class name.

    Adds module to global registry dict

    Args:
        cls: NN module class

    Raises:
        RepeatedModuleNameError: If `cls` is already in registry
    """
    try:
        MODULES[cls.__name__] = cls
    except KeyError:
        raise RepeatedModuleNameError(cls)  # pylint:disable=raise-missing-from

    return cls


def get_module(obs_space: Space, action_space: Space, config: dict) -> nn.Module:
    """Retrieve and construct module of given name.

    Args:
        obs_space: Observation space
        action_space: Action space
        config: Configurations for module construction and initialization
    """
    type_ = config.pop("type")
    if type_ not in MODULES:
        raise UnknownModuleError(type_)

    cls = MODULES[type_]
    spec = cls.spec_cls.from_dict(config)
    return cls(obs_space, action_space, spec)


for _cls in (
    DDPG,
    MAGE,
    NAF,
    SAC,
    SOP,
    ModelBasedDDPG,
    ModelBasedSAC,
    SVG,
    SoftSVG,
    TD3,
    TRPO,
):
    register(_cls)



"""Partially observed variant of the CartPole gym environment.

https://github.com/openai/gym/blob/master/gym/envs/classic_control/cartpole.py

We delete the velocity component of the state, so that it can only be solved
by an LSTM policy.
"""
import gym
import numpy as np
from gym import spaces
from gym.envs.classic_control.cartpole import CartPoleEnv


class CartPoleStateless(gym.ObservationWrapper):
    """Partially observed variant of the CartPole gym environment."""

    def __init__(self):
        super().__init__(CartPoleEnv)
        high = np.r_[
            self.env.observation_space.high[0], self.env.observation_space.high[2]
        ]
        self.observation_space = spaces.Box(-high, high, dtype=np.float32)

    def observation(self, observation):
        return np.r_[observation[0], observation[2]]



"""Learning Continuous Control Policies by Stochastic Value Gradients.

http://papers.nips.cc/paper/5796-learning-continuous-control-policies-by-stochastic-value-gradients
"""
from .policy import SVGTorchPolicy



"""Policy for MBPO using PyTorch."""
from typing import List, Tuple

from nnrl.optim import build_optimizer
from ray.rllib import SampleBatch

from raylab.agents.sac import SACTorchPolicy
from raylab.options import configure, option
from raylab.policy import learner_stats
from raylab.policy.action_dist import WrapStochasticPolicy
from raylab.policy.losses import MaximumLikelihood
from raylab.policy.model_based import (
    EnvFnMixin,
    LightningModelTrainer,
    ModelSamplingMixin,
)
from raylab.policy.model_based.policy import MBPolicyMixin, model_based_options
from raylab.policy.model_based.sampling import SamplingSpec
from raylab.utils.replay_buffer import NumpyReplayBuffer
from raylab.utils.timer import TimerStat
from raylab.utils.types import StatDict

DEFAULT_MODULE = {
    "type": "ModelBasedSAC",
    "model": {
        "network": {"units": (128, 128), "activation": "Swish"},
        "ensemble_size": 7,
        "input_dependent_scale": True,
        "parallelize": True,
        "residual": True,
    },
    "actor": {
        "encoder": {"units": (128, 128), "activation": "Swish"},
        "input_dependent_scale": True,
    },
    "critic": {
        "double_q": True,
        "encoder": {"units": (128, 128), "activation": "Swish"},
    },
    "entropy": {"initial_alpha": 0.05},
}


@configure
@model_based_options
@LightningModelTrainer.add_options
@option(
    "virtual_buffer_size",
    default=int(1e6),
    help="Size of the buffer for virtual samples",
)
@option(
    "model_rollouts",
    default=40,
    help="""Number of model rollouts to add to virtual buffer each policy interval.

    Populates virtual replay with this many model rollouts before each policy
    improvement.
    """,
)
@option(
    "real_data_ratio",
    default=0.1,
    help="Fraction of each policy minibatch to sample from environment replay pool",
)
@option("module", default=DEFAULT_MODULE, override=True)
@option(
    "optimizer/models",
    default={"type": "Adam", "lr": 3e-4, "weight_decay": 0.0001},
)
@option("model_sampling", default=SamplingSpec().to_dict(), help=SamplingSpec.__doc__)
class MBPOTorchPolicy(MBPolicyMixin, EnvFnMixin, ModelSamplingMixin, SACTorchPolicy):
    """Model-Based Policy Optimization policy in PyTorch to use with RLlib."""

    # pylint:disable=too-many-ancestors
    virtual_replay: NumpyReplayBuffer
    model_trainer: LightningModelTrainer
    dist_class = WrapStochasticPolicy

    def __init__(self, observation_space, action_space, config):
        super().__init__(observation_space, action_space, config)
        models = self.module.models
        self.loss_model = MaximumLikelihood(models)

        self.build_timers()
        self.model_trainer = LightningModelTrainer(
            models=self.module.models,
            loss_fn=self.loss_model,
            optimizer=self.optimizers["models"],
            replay=self.replay,
            config=self.config,
        )

    def _make_optimizers(self):
        optimizers = super()._make_optimizers()
        config = self.config["optimizer"]
        optimizers["models"] = build_optimizer(self.module.models, config["models"])
        return optimizers

    def build_replay_buffer(self):
        super().build_replay_buffer()
        self.virtual_replay = NumpyReplayBuffer(
            self.observation_space,
            self.action_space,
            self.config["virtual_buffer_size"],
        )
        self.virtual_replay.seed(self.config["seed"])

    def build_timers(self):
        super().build_timers()
        self.timers["augmentation"] = TimerStat()

    @learner_stats
    def learn_on_batch(self, samples: SampleBatch) -> dict:
        self.add_to_buffer(samples)
        self._learn_calls += 1

        info = {}
        warmup = self._learn_calls == 1
        if self._learn_calls % self.config["model_update_interval"] == 0 or warmup:
            with self.timers["model"] as timer:
                losses, model_info = self.train_dynamics_model(warmup=warmup)
                timer.push_units_processed(model_info["model_epochs"])
                info.update(model_info)
            self.set_new_elite(losses)

        with self.timers["augmentation"] as timer:
            count_before = len(self.virtual_replay)
            self.populate_virtual_buffer()
            timer.push_units_processed(len(self.virtual_replay) - count_before)

        with self.timers["policy"] as timer:
            times = self.config["improvement_steps"]
            policy_info = self.update_policy(times=times)
            timer.push_units_processed(times)
            info.update(policy_info)

        info.update(self.timer_stats())
        return info

    def train_dynamics_model(
        self, warmup: bool = False
    ) -> Tuple[List[float], StatDict]:
        return self.model_trainer.optimize(warmup=warmup)

    def populate_virtual_buffer(self):
        # pylint:disable=missing-function-docstring
        num_rollouts = self.config["model_rollouts"]
        real_data_ratio = self.config["real_data_ratio"]
        if not (num_rollouts and real_data_ratio < 1.0):
            return

        real_samples = self.replay.sample(num_rollouts)
        virtual_samples = self.generate_virtual_sample_batch(real_samples)
        self.virtual_replay.add(virtual_samples)

    def update_policy(self, times: int) -> StatDict:
        batch_size = self.config["batch_size"]
        env_batch_size = int(batch_size * self.config["real_data_ratio"])
        model_batch_size = batch_size - env_batch_size

        for _ in range(times):
            samples = []
            if env_batch_size:
                samples += [self.replay.sample(env_batch_size)]
            if model_batch_size:
                samples += [self.virtual_replay.sample(model_batch_size)]
            batch = SampleBatch.concat_samples(samples)
            batch = self.lazy_tensor_dict(batch)
            info = self.improve_policy(batch)

        return info

    def timer_stats(self) -> dict:
        stats = super().timer_stats()
        augmentation_timer = self.timers["augmentation"]
        stats.update(
            augmentation_time_s=round(augmentation_timer.mean, 3),
            augmentation_throughput=round(augmentation_timer.mean_throughput, 3),
        )
        return stats



"""
From Proximal Policy Optimization Algorithms:
http://arxiv.org/abs/1707.06347
"""
from dataclasses import dataclass, field


@dataclass
class AdaptiveKLCoeffSpec:
    """Adaptive schedule for KL Divergence regularization."""

    initial_coeff: float = 0.2
    desired_kl: float = 0.01
    adaptation_coeff: float = 2.0
    threshold: float = 1.5
    curr_coeff: float = field(init=False)

    def __post_init__(self):
        self.curr_coeff = self.initial_coeff

    def adapt(self, kl_div):
        """Apply PPO rule to update current KL coeff based on latest KL divergence."""
        if kl_div < self.desired_kl / self.threshold:
            self.curr_coeff /= self.adaptation_coeff
        elif kl_div > self.desired_kl * self.threshold:
            self.curr_coeff *= self.adaptation_coeff



"""Network and configurations for model-based SAC algorithms."""
from dataclasses import dataclass, field

from gym.spaces import Box
from nnrl.nn.model import EnsembleSpec, build_ensemble

from .sac import SAC, SACSpec


@dataclass
class ModelBasedSACSpec(SACSpec):
    """Specifications for model-based SAC modules.

    Args:
        model: Specifications for stochastic dynamics model ensemble
        actor: Specifications for stochastic policy and entropy coefficient
        critic: Specifications for action-value estimators
        initializer: Optional dictionary with mandatory `type` key corresponding
            to the initializer function name in `torch.nn.init` and optional
            keyword arguments. Overrides model, actor, and critic initializer
            specifications.
    """

    model: EnsembleSpec = field(default_factory=EnsembleSpec)

    def __post_init__(self):
        super().__post_init__()
        if self.initializer:
            self.model.initializer = self.initializer


class ModelBasedSAC(SAC):
    """NN module for Model-Based Soft Actor-Critic algorithms.

    Args:
        obs_space: Observation space
        action_space: Action space
        spec: Specifications for model-based SAC modules

    Attributes:
        models (SME): Stochastic dynamics model ensemble
        actor (StochasticPolicy): Stochastic policy to be learned
        alpha (Alpha): Entropy bonus coefficient
        critics (QValueEnsemble): The action-value estimators to be learned
        target_critics (QValueEnsemble): The action-value estimators used for
            bootstrapping in Q-Learning
        spec_cls: Expected class of `spec` init argument
    """

    # pylint:disable=abstract-method
    spec_cls = ModelBasedSACSpec

    def __init__(self, obs_space: Box, action_space: Box, spec: ModelBasedSACSpec):
        super().__init__(obs_space, action_space, spec)

        self.models = build_ensemble(obs_space, action_space, spec.model)



"""Generic Trainer and base configuration for model-based agents."""
import logging
from typing import Any

from ray.rllib import Policy
from ray.rllib.evaluation.worker_set import WorkerSet

logger = logging.getLogger(__name__)


def set_policy_with_env_fn(worker_set: WorkerSet, fn_type: str = "reward"):
    """Set the desired environment function for all policies in the worker set.

    Args:
        worker_set: A worker set instance, usually from a trainer
        fn_type: The type of environment function, either 'reward',
            'termination', or 'dynamics'
        from_env: Whether to retrieve the function from the environment instance
            or from the global registry
    """
    worker_set.foreach_worker(
        lambda w: w.foreach_policy(
            lambda p, _: _set_from_env_if_possible(p, w.env, fn_type)
        )
    )


def _set_from_env_if_possible(policy: Policy, env: Any, fn_type: str = "reward"):
    env_fn = getattr(env, fn_type + "_fn", None)
    if fn_type == "reward":
        if env_fn:
            policy.set_reward_from_callable(env_fn)
        else:
            policy.set_reward_from_config()
    elif fn_type == "termination":
        if env_fn:
            policy.set_termination_from_callable(env_fn)
        else:
            policy.set_termination_from_config()
    elif fn_type == "dynamics":
        if env_fn:
            policy.set_dynamics_from_callable(env_fn)
        else:
            raise ValueError(
                f"Environment '{env}' has no '{fn_type + '_fn'}' attribute"
            )
    else:
        raise ValueError(f"Invalid env function type '{fn_type}'")


class ModelBasedMixin:
    """Mixin for model-based agents.

    Sets reward and termination functions for policies. These functions must be
    either:
    * Registered via `raylab.envs.register_reward_fn` and
      `raylab.envs.register_termination_fn`
    * Accessible attributes of the environment as `reward_fn` and
      `termination_fn`. These should not be bound instance methods; all
      necessary information should be encoded in the inputs, (state, action,
      and next state) i.e., the states should be markovian.
    """

    def after_init(self):
        # pylint:disable=missing-function-docstring
        super().after_init()
        self.set_env_fns()

    def set_env_fns(self):
        """Set reward and termination functions for policies."""
        set_policy_with_env_fn(self.workers, fn_type="reward")
        set_policy_with_env_fn(self.workers, fn_type="termination")



"""Debugging utilities."""
import numpy as np
from gym import spaces
from ray.rllib import SampleBatch


def fake_space_samples(space, batch_size):
    """Create fake samples from a Gym space."""
    if isinstance(space, spaces.Box):
        shape = (batch_size,) + space.shape
        randn = np.random.randn(*shape)
        return np.clip(randn, space.low, space.high).astype(space.dtype)
    if isinstance(space, spaces.Discrete):
        randint = np.random.randint(space.n, size=batch_size)
        return randint.astype(space.dtype)
    raise ValueError(f"Unsupported space type {type(space)}")


def fake_batch(obs_space, action_space, batch_size=1):
    """Create a fake SampleBatch compatible with Policy.learn_on_batch."""
    samples = {
        SampleBatch.CUR_OBS: fake_space_samples(obs_space, batch_size),
        SampleBatch.ACTIONS: fake_space_samples(action_space, batch_size),
        SampleBatch.REWARDS: np.random.randn(batch_size).astype(np.float32),
        SampleBatch.NEXT_OBS: fake_space_samples(obs_space, batch_size),
        SampleBatch.DONES: np.random.randn(batch_size) > 0,
    }
    return SampleBatch(samples)



"""Registry of environment termination functions in PyTorch."""
import torch
from torch import nn

from raylab.tune.registry import RAYLAB_TERMINATION, _raylab_registry

from .utils import get_env_parameters, has_env_creator

# For testing purposes
TERMINATIONS = {}


def has_termination_fn(env_id: str) -> bool:
    """Whether a termination function for the env id is in the global registry."""
    return _raylab_registry.contains(RAYLAB_TERMINATION, env_id)


def get_termination_fn(env_id, env_config=None):
    """Return the termination funtion for the given environment name and configuration.

    Only returns for environments which have been registered with Tune.
    """
    assert has_env_creator(env_id), f"{env_id} environment not registered with Tune."
    assert has_termination_fn(
        env_id
    ), f"{env_id} environment termination not registered."

    env_config = env_config or {}
    termination_fn = _raylab_registry.get(RAYLAB_TERMINATION, env_id)(env_config)
    if env_config.get("time_aware", False):
        termination_fn = TimeAwareTerminationFn(termination_fn)
    return termination_fn


def register(*ids):
    """Register termination function class for environments with given ids."""

    def librarian(cls):
        for id_ in ids:
            TERMINATIONS[id_] = cls
            _raylab_registry.register(RAYLAB_TERMINATION, id_, cls)

        return cls

    return librarian


class TerminationFn(nn.Module):
    """
    Module that computes an environment's termination function for batches of inputs.
    """

    def forward(self, state, action, next_state):  # pylint:disable=arguments-differ
        raise NotImplementedError


class TimeAwareTerminationFn(TerminationFn):
    """Wraps a termination function and removes time dimension before forwarding."""

    def __init__(self, termination_fn):
        super().__init__()
        self.termination_fn = termination_fn

    def forward(self, state, action, next_state):
        timeout = next_state[..., -1] >= 1.0
        env_done = self.termination_fn(state[..., :-1], action, next_state[..., :-1])
        return timeout | env_done


################################################################################
# Built-ins
################################################################################


@register(
    "HalfCheetah-v2",
    "HalfCheetah-v3",
    "IndustrialBenchmark-v0",
    "IBOperationalCost-v0",
    "IBMisCalibration-v0",
    "IBFatigue-v0",
    "ReacherBulletEnv-v0",
    "Pusher-v2",
    "Swimmer-v2",
    "Swimmer-v3",
    "Pendulum-v0",
)
class NoTermination(TerminationFn):
    """Termination function for continuing environments.

    This is usually the case for environments which need an external time limit
    enforcement. In that situation, wrap this with TimeAwareTerminationFn.
    """

    def __init__(self, _):
        super().__init__()

    def forward(self, state, action, next_state):
        return torch.zeros(next_state.shape[:-1]).bool()


@register(
    "CartPoleSwingUp-v0",
    "CartPoleSwingUp-v1",
    "TorchCartPoleSwingUp-v0",
    "TorchCartPoleSwingUp-v1",
)
class CartPoleSwingUpTermination(TerminationFn):
    """CartPoleSwingUp's termination function."""

    def __init__(self, _):
        from gym_cartpole_swingup.envs.cartpole_swingup import CartPoleSwingUpParams

        super().__init__()
        params = CartPoleSwingUpParams()
        self.x_threshold = params.x_threshold

    def forward(self, state, action, next_state):
        return next_state[..., 0].abs() > self.x_threshold


@register("MountainCarContinuous-v0")
class MountainCarContinuousTermination(TerminationFn):
    """MountainCarContinuous' reward function."""

    def __init__(self, config):
        super().__init__()
        goal_position = 0.45
        goal_velocity = config.get("goal_velocity", 0.0)
        self.goal = torch.as_tensor([goal_position, goal_velocity])

    def forward(self, state, action, next_state):
        return (next_state >= self.goal).all(dim=-1)


@register("HVAC")
class HVACTermination(TerminationFn):
    """HVAC's termination function."""

    def __init__(self, _):
        super().__init__()

    def forward(self, state, action, next_state):
        return next_state[..., -1] >= 1.0


@register("Reservoir")
class ReservoirTermination(TerminationFn):
    """Reservoir's termination function."""

    def __init__(self, _):
        super().__init__()

    def forward(self, state, action, next_state):
        return next_state[..., -1] >= 1.0


@register("Navigation")
class NavigationTermination(TerminationFn):
    """Navigation's termination function."""

    def __init__(self, config):
        from .environments.navigation import DEFAULT_CONFIG

        super().__init__()
        config = {**DEFAULT_CONFIG, **config}
        self.end = torch.as_tensor(config["end"]).float()

    def forward(self, state, action, next_state):
        hit_goal = ((next_state[..., :2] - self.end).abs() <= 1e-1).all(dim=-1)
        timeout = next_state[..., -1] >= 1
        return hit_goal | timeout


# @register("Pendulum-v0")
# class PendulumTermination(TerminationFn):
#     """Pendulum-v0's termination function."""


@register("Walker2d-v3")
class Walker2DTermination(TerminationFn):
    """Walker2d-v3's termination function."""

    def __init__(self, config):
        super().__init__()
        parameters = get_env_parameters("Walker2d-v3")
        for attr in """
        terminate_when_unhealthy
        healthy_z_range
        healthy_angle_range
        exclude_current_positions_from_observation
        """.split():
            setattr(self, "_" + attr, config.get(attr, parameters[attr].default))

    def _is_healthy(self, state):
        # pylint:disable=invalid-name
        if self._exclude_current_positions_from_observation:
            z, angle = state[..., 0], state[..., 1]
        else:
            z, angle = state[..., 1], state[..., 2]

        min_z, max_z = self._healthy_z_range
        min_angle, max_angle = self._healthy_angle_range

        healthy_z = (min_z < z) & (z < max_z)
        healthy_angle = (min_angle < angle) & (angle < max_angle)
        is_healthy = healthy_z & healthy_angle

        return is_healthy

    def forward(self, state, action, next_state):
        if self._terminate_when_unhealthy:
            return ~self._is_healthy(next_state)
        return torch.zeros(next_state.shape[:-1]).bool()


@register("Hopper-v3")
class HopperTermination(TerminationFn):
    """Hopper-v3's termination function."""

    def __init__(self, config: dict):
        super().__init__()
        parameters = get_env_parameters("Hopper-v3")
        for attr in """
        exclude_current_positions_from_observation
        healthy_angle_range
        healthy_state_range
        healthy_z_range
        terminate_when_unhealthy
        """.split():
            setattr(self, "_" + attr, config.get(attr, parameters[attr].default))

    def forward(self, state, action, next_state):
        if self._terminate_when_unhealthy:
            return ~self.is_healthy(next_state)
        return torch.zeros(next_state.shape[:-1]).bool()

    def is_healthy(self, state):
        # pylint:disable=invalid-name
        if self._exclude_current_positions_from_observation:
            z, angle, state_ = state[..., 0], state[..., 1], state[..., 1:]
        else:
            z, angle, state_ = state[..., 1], state[..., 2], state[..., 2:]

        min_state, max_state = self._healthy_state_range
        min_z, max_z = self._healthy_z_range
        min_angle, max_angle = self._healthy_angle_range

        healthy_state = torch.all((min_state < state_) & (state_ < max_state), dim=-1)
        healthy_z = (min_z < z) & (z < max_z)
        healthy_angle = (min_angle < angle) & (angle < max_angle)

        return healthy_state & healthy_z & healthy_angle


@register("InvertedPendulum-v2")
class InvertedPendulumTermination(TerminationFn):
    # pylint:disable=abstract-method,missing-class-docstring
    def __init__(self, _):
        super().__init__()

    def forward(self, state, action, next_state):
        # pylint:disable=no-self-use
        notdone = torch.isfinite(next_state).all(dim=-1) & (
            next_state[..., 1].abs() <= 0.2
        )
        return ~notdone
